
/* z0 simulator

   25 Aug 2014
   P. Chongstitvatana
   Department of Computer Engineering
   Chulalongkorn University

*/

#include "z0.h"

int R[16], IR, PC, AC, F, M[MAXMEM];
int runflag, savePc;

int stk[10], sp;		// return stack, stack pointer

void push(int a){
	sp++;
	if( sp > 9 ) error("nested call overflow");
	stk[sp] = a;
}

int pop(void){
	int a;
	a = stk[sp];
	sp--;
	if( sp < 0 ) error("nested call underflow");
	return a;
}

void run(void){		// execute one instruction
	int op, arg;
	savePc = PC;
	IR = M[PC];		// fetch instruction
	PC++;
	op = IRop();	// op is leftmost 8 bits
	arg = IRarg();	// arg is rightmost 24 bits
	ninst++;
	switch( op ) {
	case ADD:  AC += R[arg]; break;
	case SUB:  AC -= R[arg]; break;
	case INC:  R[arg]++; break;
	case DEC:  R[arg]--; break;
	case MOV:  AC = R[arg]; break;
	case MVI:  AC = arg; break;
	case PUT:  R[arg] = AC; break;
	case EQ:   F = AC == R[arg]; break;
	case NE:   F = AC != R[arg]; break;
	case LT:   F = AC < R[arg]; break;
	case LE:   F = AC <= R[arg]; break;
	case GT:   F = AC > R[arg]; break;
	case GE:   F = AC >= R[arg]; break;
	case Z:    F = R[arg] == 0; break;
	case JMP:  PC = arg; break;
	case JT:   if ( F != 0 ) PC = arg; break;
	case JF:   if ( F == 0 ) PC = arg; break;
	case STOP: runflag = 0; break;
	case LD:   AC = M[arg]; break;
	case ST:   M[arg] = AC; break;
	case LDD:  AC = M[R[arg]]; break;
	case STD:  M[R[arg]] = AC; break;
	case CALL: push(PC); PC = arg; break;
	case RET:  PC = pop(); break;
	default: error("undefine op");
	}
}

void run2(void){
	if( runflag ) run();
}

int main(int argc, char *argv[]){
	if( argc < 2 ) {
		printf("usage : z0 objfile\n");
		exit(0);
	}
	loadprogram(argv[1]);
	ninst = 0;
	PC = 0;
	F = 0;
	sp = 0;
	runflag = 1;
	interp();
	return 0;
}

